package ida.pelab.prt.generalfunctions;


import ida.pelab.prt.transformation.ExpressionSimplifier;

import java.util.ArrayList;

import cetus.hir.ArrayAccess;
import cetus.hir.AssignmentExpression;

import cetus.hir.DepthFirstIterator;
import cetus.hir.Expression;

import cetus.hir.CompoundStatement;

import cetus.hir.ForLoop;
import cetus.hir.IntegerLiteral;
import cetus.hir.PostOrderIterator;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;

public class TraverseFunction {
	
	
	

	

	public static boolean hasChildren(Traversable node)
	{
				return node.getChildren()!=null && node.getChildren().size()>0;
		
	}
	
	public static ArrayAccess  GetArray(Traversable  object)
	{
			
		if(object instanceof ArrayAccess)
			return (ArrayAccess)object;

		else return null ;
	
	}
	
	public static boolean hasSimpleIndices(Expression indexVariable,Traversable node)
	{
		int  indexVariableCounter=0;
		PostOrderIterator<Traversable>  customIterator = new PostOrderIterator<Traversable>(node);
		while(customIterator.hasNext())
		{
			Object currentObject =customIterator.next();
			
			if((currentObject instanceof Expression)&& ((Expression)currentObject).equals(indexVariable) )
				indexVariableCounter++;
			
			// check whether current object is and array which does not contain i as part of the indices
			if((currentObject instanceof ArrayAccess)&& !((Traversable)currentObject).equals(node) && ExpressionContainsVariable(indexVariable, (Traversable)currentObject ))
			return false;
		}
		
		return indexVariableCounter <2 ;
		
		
		
	}
	
	
	
	//checks whether a certain variable occurs in the indices of an array
	
	public static boolean  ExpressionContainsVariable(Expression indexVariable,Traversable exprNode)
	{
		
		PostOrderIterator<Traversable>  customIterator = new PostOrderIterator<Traversable>(exprNode);
		while(customIterator.hasNext())
		{
			Object currentObject =customIterator.next();
			if((currentObject instanceof Expression)&& ((Expression)currentObject).equals(indexVariable) )
				return true ;
			
		
			
		}
		return false;
	}
	
	

	public static ArrayList<Integer> findSubscriptedIndices(ArrayAccess array, Expression indexVariable)
	{
		ArrayList<Integer> indicesArray = new ArrayList<Integer>(array.getIndices().size()); 
		for(int i=0;i< array.getIndices().size();i++)
		{
			 Expression indexExpr= ExpressionSimplifier.Simplify(array.getIndex(i));
			 if(isOccurenceValid(indexExpr,indexVariable))
				 indicesArray.add(i);			 		
		}
		return indicesArray;
	}
	
	
	
	private static boolean  isOccurenceValid(Traversable expr,Expression indexVariable)
	{
		
		int counter=0;
		PostOrderIterator<Traversable>  customIterator = new PostOrderIterator<Traversable>(expr);
		while(customIterator.hasNext())
		{
			Object currentObject =customIterator.next();
			if((currentObject instanceof Expression)&& ((Expression)currentObject).equals(indexVariable) )
				{
				   Expression  cuObj=(Expression)currentObject;
				   //invalidcase
				   if(cuObj instanceof ArrayAccess)
					   counter+=5;
				   else
					   counter++;
					   
				}
			
		}
		
		return counter==1;
	}
	
	
	
	
	
	
	//checks whether the input array has an index in the form presented by the indexExpr parameter
	public static boolean ContainsSpecifiedIndex(ArrayAccess array, Expression indexExpr)
	{
		
		for(int i= 0; i<array.getIndices().size();i++)
		    if(array.getIndex(i).equals(indexExpr))
				return true;
			
		return false ;
		
	}
	
	//checks whether the input array has an index which is equivalent to the form presented by the indexExpr parameter
	public static boolean ContainsEquivalentIndex(ArrayAccess array, Expression indexExpr)
	{
		
		
		for(int i= 0; i<array.getIndices().size();i++)
		    if(ExpressionSimplifier.Simplify( array.getIndex(i)).equals(ExpressionSimplifier.Simplify( indexExpr)))
				return true;
			
		return false ;
		
	}
	
	@Deprecated
	public static boolean isReadorWritten(Traversable key ,Traversable LastNode)
	{
		
		Traversable startPoint = key.getParent().getParent().getParent();
		
		DepthFirstIterator<Traversable> iterator = new DepthFirstIterator<Traversable>(startPoint);
		int count = 0;
		boolean  reachedkey= false ;
		 while(iterator.hasNext())
		 {
			 
			 Traversable nextNode =(Traversable) iterator.next();		
			 
			 if(System.identityHashCode(nextNode)==System.identityHashCode(key))
				 reachedkey=true ;
			 
			 if(reachedkey)
			 {
			 if( nextNode.equals(key))
				 count++;
			 
			 if(System.identityHashCode(nextNode)==System.identityHashCode(LastNode))
				 break;
				 
			 }
			 
		 }
		
		
	return count==1 ;	
	}

	
	
	public static boolean isWritten(Traversable startPoint ,Traversable endPoint,Traversable key)
	{
		Traversable compoundStart= getTopLevelCompoundStatement(startPoint);
		Traversable startingSibling = getNextSibling(getTopLevelNode(startPoint));
		if(compoundStart==null || startingSibling==null)
			 return false;
		if(System.identityHashCode(startingSibling)==System.identityHashCode(endPoint))
			return false;
		
		DepthFirstIterator<Traversable> iterator = new DepthFirstIterator<Traversable>(compoundStart);
		int count = 0;
		boolean  reachedPoint= false ;
		 while(iterator.hasNext())
		 { 
			 Traversable nextNode =iterator.next();		
			 if(reachedPoint)
			 {
				 if( nextNode.equals(key))	
				 {
					 if(nextNode.getParent() instanceof AssignmentExpression && ((AssignmentExpression)nextNode.getParent()).getLHS().equals((Expression)key )) 
						 count++;
					 else if (nextNode.getParent() instanceof UnaryExpression &&  isValidOperator( ((UnaryExpression)nextNode.getParent()).getOperator()) )
						 count++;
				 }
				 	
				 if(System.identityHashCode(nextNode)==System.identityHashCode(endPoint))
					 break;
				 
			 }
			 else if(System.identityHashCode(nextNode)==System.identityHashCode(startingSibling))
				  reachedPoint=true ;
			
			 
		 }
		
		
	return count>0 ;	
	}
	


	
	public static boolean isRead(Traversable startPoint ,Traversable endPoint,Traversable key)
	{
		Traversable compoundStart= getTopLevelCompoundStatement(startPoint);
		Traversable srchStartPoint = getNextSibling(getTopLevelNode(startPoint));
		if(compoundStart==null ||srchStartPoint==endPoint)
			 return false;
		if(System.identityHashCode(srchStartPoint)==System.identityHashCode(endPoint))
			return false;
		DepthFirstIterator<Traversable> iterator = new DepthFirstIterator<Traversable>(compoundStart);
		int count = 0;
		boolean  reachedPoint= false ;		
		 while(iterator.hasNext())
		 {
			 
			 Traversable nextNode =iterator.next();		
			 if(reachedPoint)
			 {
				 if( nextNode.equals(key))	
				 {
					 count++;
					 //write accesses
					 if( nextNode.getParent() instanceof AssignmentExpression && 
						((AssignmentExpression)nextNode.getParent()).getLHS().equals((Expression)key )) 
						  count--;
				 }
				 	
				 if(System.identityHashCode(nextNode)==System.identityHashCode(endPoint))
					 break;
				 
			 }
			 else if(System.identityHashCode(nextNode)==System.identityHashCode(srchStartPoint))
				  reachedPoint=true ;
			 
		 }
		
		
	return count>0 ;	
	}
	
	
	
	
	
	
	
	
	
	
	private static Traversable getNextSibling(Traversable node)
	{
		Traversable nextSibling = null;
		
		if( node!=null && node.getParent()!=null)
		{
			Traversable parentNode=node.getParent();
			
			int nodeIndex=-1;
			for(int j=0; j< parentNode.getChildren().size();j++)
				if(System.identityHashCode(node)==System.identityHashCode(parentNode.getChildren().get(j)))
						{
							nodeIndex=j;
							break;
						}
			if(nodeIndex>-1 && nodeIndex< parentNode.getChildren().size()-1)
				nextSibling = parentNode.getChildren().get(nodeIndex+1);
		}
		
		return nextSibling;
	}
	
	private static Traversable getTopLevelCompoundStatement(Traversable node)
	{
		Traversable result;
		for(result=node.getParent(); result!=null &&  !(result instanceof CompoundStatement );result=result.getParent() );
		return result;	
		
		
	}
	
	private static Traversable getTopLevelNode(Traversable node)
	{
		
		if(node instanceof Expression)
			return node.getParent();
		else return node;
		
	}
	
	
	
	
	
	
	
	
	
	
	@Deprecated
	public static boolean isWritten(Traversable key ,Traversable LastNode)
	{
		
		Traversable startPoint = key.getParent().getParent().getParent();
		
		DepthFirstIterator<Traversable> iterator = new DepthFirstIterator<Traversable>(startPoint);
		int count = 0;
		boolean  reachedkey= false ;
		 while(iterator.hasNext())
		 {
			 
			 Traversable nextNode =(Traversable) iterator.next();		
			 
			 if(System.identityHashCode(nextNode)==System.identityHashCode(key))
				 reachedkey=true ;
			 
			 else if(reachedkey)
			 {
				 if( nextNode.equals(key))	
				 {
					 if(nextNode.getParent() instanceof AssignmentExpression && ((AssignmentExpression)nextNode.getParent()).getLHS().equals((Expression)key )) 
						 count++;
					 else if (nextNode.getParent() instanceof UnaryExpression &&  isValidOperator( ((UnaryExpression)nextNode.getParent()).getOperator()) )
						 count++;
				 }
				 	
				 if(System.identityHashCode(nextNode)==System.identityHashCode(LastNode))
					 break;
				 
			 }
			 
		 }
		
		
	return count>1 ;	
	}
	
	
	
	private static boolean isValidOperator(UnaryOperator up)
	{
		return up==UnaryOperator.POST_DECREMENT || up==UnaryOperator.PRE_DECREMENT || up==UnaryOperator.POST_INCREMENT || up==UnaryOperator.PRE_INCREMENT;
	}
	
	// checks the existence of node in the subtree of another node 
	public static boolean  subTreeContainsExpression(Traversable searchToken,Traversable subtreeRoot)
	{
		
		PostOrderIterator<Traversable>  customIterator = new PostOrderIterator<Traversable>(subtreeRoot);
		while(customIterator.hasNext())
		{
			Object currentObject =customIterator.next();
			if(currentObject.equals(searchToken) )
				return true ;
		}
		return false;
	}
	

	
	
	// Calculates the domain of memory accesses by an array inside a certain for loop
	// supports only one dimensional arrays
	public static ArrayList<String>  calculateMemoryAccessDomain(ForLoop parentLoop,ArrayAccess array)
	{
		//ForLoop  forLoop = (ForLoop)parentLoop;
		
	//	Expression lb=  LoopFunctions.getLowerBoundExpression(forLoop);
		//Expression ub= LoopFunctions.getUpperBoundExpression(forLoop);
		//Expression loopVariableExpr = LoopFunctions.getLoopvariable(forLoop);
		
		//Expression step = LoopFunctions.getStepIncrement(forLoop);
	 
		ArrayList<String>  resultSet = new ArrayList<String>();
		
		
		return resultSet;
		
	}
	
	
	
	public static  boolean compareRange(Expression variable,int value)
	{
		IntegerLiteral valueLiteral = new IntegerLiteral(value);
		return ExpressionSimplifier.Simplify(variable).equals(valueLiteral);
		

	}
	
	
	
	
	
}
