package ida.pelab.prt.pattern;




import ida.pelab.prt.generalfunctions.LoopDirectionType;
import ida.pelab.prt.generalfunctions.LoopFunctions;
import ida.pelab.prt.transformation.ExpressionSimplifier;







import cetus.hir.BinaryExpression;
import cetus.hir.BinaryOperator;
import cetus.hir.Expression;
import cetus.hir.ForLoop;
import cetus.hir.IntegerLiteral;



public class LoopVariable extends Expression{
	
	private Expression lowerBound ;
	private Expression upperBound ;
	private Expression step;
	private Expression  variableExp;
	private LoopDirectionType loopDirection;
	private BinaryOperator ConditionOpr;
	
	public LoopVariable(){};
	public LoopVariable(Expression ex,Expression lb,Expression ub,Expression step,LoopDirectionType dir,BinaryOperator op )
	{
		setAnnotatedExpression(ex);
		setLowerBound(lb);
		setUpperBound(ub);
		setStep(step);
		setDirection(dir);
		setConditionOpr(op);
		
	}
	public void setDirection(LoopDirectionType dir) {
		this.loopDirection = dir;
	}
	public LoopDirectionType getDirection() {
		return loopDirection;
	}
	public void setLowerBound(Expression lowerBound) {
		this.lowerBound = lowerBound;
	}
	public Expression getLowerBound() {
		return lowerBound;
	}
	public void setStep(Expression step) {
		this.step = step;
	}
	public Expression getStep() {
		return step;
	}
	public void setUpperBound(Expression upperBound) {
		this.upperBound = upperBound;
	}
	public Expression getUpperBound() {
		return upperBound;
	}
	
	public Expression getAbsoluteUpperBound() {
		
		Expression result=upperBound;
			if(ConditionOpr.equals(BinaryOperator.COMPARE_LT))
				result =new BinaryExpression(upperBound, BinaryOperator.ADD, new IntegerLiteral(1));
			else if (ConditionOpr.equals(BinaryOperator.COMPARE_GT))
				result =new BinaryExpression(upperBound, BinaryOperator.SUBTRACT, new IntegerLiteral(1));
			if(!result.equals(upperBound))
				result= ExpressionSimplifier.Simplify( result);
		return result;			
	}

	public void setAnnotatedExpression(Expression expression) {
		this.variableExp = expression;
	}
	public Expression getVariableExpression() {
		return variableExp;
	}
	
	
	public static LoopVariable buildVariable(ForLoop loop)
	{
		Expression loopVariableExpr = LoopFunctions.getLoopvariable(loop);
		Expression lb=LoopFunctions.getLowerBoundExpression(loop);
		Expression ub=LoopFunctions.getUpperBoundExpression(loop);
		LoopDirectionType direction = LoopFunctions.detectLoopDirection(loop);
		Expression step =LoopFunctions.getStepIncrement(loop);
		BinaryOperator op =LoopFunctions.getLoopOperator(loop);
		return new LoopVariable(loopVariableExpr, lb, ub, step, direction,op);
	
	}
	
	
	public  String toString()
	{
		String resultString=	ExpressionSimplifier.Simplify(lowerBound).toString()+":"+	ExpressionSimplifier.Simplify(upperBound).toString() ;
		if(step!=null && !ExpressionSimplifier.Simplify(step).toString().equals("1"))
			resultString+=":"+ExpressionSimplifier.Simplify(step).toString();
		return resultString ;
		
	}
	

	public  String toAnnotationString()
	{
		String resultString=	ExpressionSimplifier.Simplify(lowerBound).toString()+":"+	ExpressionSimplifier.Simplify(upperBound).toString() ;
		return resultString ;
		
	}
	public String toIndexString()
	{
		return 	ExpressionSimplifier.Simplify(variableExp).toString()+"=[" + toString()+ "]";
	}
	
	
	
	public boolean  equals(Object input)
	{
		if(this==input)
			return true ;
		else if(input instanceof Expression)
			return variableExp.equals((Expression)input);
		
		return false;
	}
	
	
	public static boolean SymbolicComparison(LoopVariable first,LoopVariable second)
	{
		return ExpressionSimplifier.Simplify(first.getVariableExpression()).equals(ExpressionSimplifier.Simplify(second.getVariableExpression()))
		&&
		ExpressionSimplifier.Simplify(first.getLowerBound()).equals(ExpressionSimplifier.Simplify(second.getLowerBound()))
		&&
		ExpressionSimplifier.Simplify(first.getUpperBound()).equals(ExpressionSimplifier.Simplify(second.getUpperBound()))
		&&
		ExpressionSimplifier.Simplify(first.getStep()).equals(ExpressionSimplifier.Simplify(second.getStep()));		
		
	}
	public void setConditionOpr(BinaryOperator conditionOpr) {
		ConditionOpr = conditionOpr;
	}
	public BinaryOperator getConditionOpr() {
		return ConditionOpr;
	}
	
	
	
}
