package ida.pelab.prt.pattern;





import ida.pelab.prt.constraints.Constraint;
import ida.pelab.prt.constraints.RuleParameterType;
import ida.pelab.prt.constraints.TypeCheck;

import java.util.ArrayList;




/**
 * This class represents the structure for the definition of each pattern.
 *
 */
public class PatternStructure {
	
	
	public PatternStructure(){ 
		helpPattern=false;
		pruneIFCondtion=false;
		}
	
	private String name ;
	private PatternType type;
	private PatternLevel level ;
	private PatternStructureOperator operator;
	private ArrayList<ArrayList<String>> structureElement ;
	private ArrayList<Constraint> rules;
	private PatternOutputFormat   outputSyntax;
	private boolean helpPattern;
	private boolean customMatchingRules;
	
	/**
	 * for skipping matched patterns in the condition of an IfStatement
	 */
	private boolean pruneIFCondtion;
	
	public PatternStructureOperator getOperator() {
		return operator;
	}
	public void setOperator(PatternStructureOperator operator) {
		this.operator = operator;
	}
	public ArrayList<ArrayList<String>> getStructureElement() {
		return structureElement;
	}
	public void setStructureElement(ArrayList<ArrayList<String>> structureElement) {
		this.structureElement = structureElement;
	}
	public ArrayList<Constraint> getConstraints() {
		return rules;
	}
	public void setConstraints(ArrayList<Constraint> rules) {
		this.rules = rules;
	}
	public PatternOutputFormat getOutputSyntax() {
		return outputSyntax;
	}
	public void setOutputSyntax(PatternOutputFormat outputSyntax) {
		this.outputSyntax = outputSyntax;
	}
	public void setName(String name) {
		this.name = name;
	}
	public String getName() {
		return name;
	}
	public void setType(PatternType type) {
		this.type = type;
	}
	public PatternType getType() {
		return type;
	}
	public void setLevel(PatternLevel level) {
		this.level = level;
	}
	public PatternLevel getLevel() {
		return level;
	}

	
	public static PatternStructure getNoPatternInstance()
	{
		PatternStructure noPattern= new PatternStructure();
		noPattern.setName("No_Pattern");
		noPattern.setLevel(PatternLevel.Trivial);
		noPattern.setType(PatternType.Vertical);
		noPattern.setOutputSyntax(new PatternOutputFormat());
		noPattern.setConstraints(new ArrayList<Constraint>());
		return noPattern;
	}
	
	
	public  RuleParameterType getType(String  stringRef)
	{
		for(int i=0 ; i< rules.size();i++)
			if(rules.get(i) instanceof TypeCheck)
			{
				TypeCheck tmpCheck =(TypeCheck)rules.get(i);
				if(tmpCheck.getParameter().equals(stringRef))
					return  tmpCheck.getParamType();
			}
		
		return RuleParameterType.Invalid;
	}
	
  public String toString()
  {
	  return name;
  }
public void setHelpPattern(boolean helpPattern) {
	this.helpPattern = helpPattern;
}
public boolean isHelpPattern() {
	return helpPattern;
}
public boolean isTrivial() {
	return level.equals(PatternLevel.Trivial);
}
public void setCustomMatchingRules(boolean customMatchingRules) {
	this.customMatchingRules = customMatchingRules;
}
public boolean hasCustomMatchingRules() {
	return customMatchingRules;
}
public void setPruneIFCondtion(boolean pruneIFCondtion) {
	this.pruneIFCondtion = pruneIFCondtion;
}
public boolean isPruneIFCondtion() {
	return pruneIFCondtion;
}

	
	

}


