package ida.pelab.prt.patternhierarchy;

import ida.pelab.prt.analysis.RecognitionAnalysis;

import ida.pelab.prt.generalfunctions.LoopFunctions;
import ida.pelab.prt.pattern.MatchedInstance;


import java.util.ArrayList;

import java.util.Hashtable;

import cetus.hir.Expression;
import cetus.hir.ExpressionStatement;
import cetus.hir.ForLoop;

import cetus.hir.Statement;
import cetus.hir.Traversable;

public class AnnotatedTree {
	
	
	
	private final static Hashtable<Integer, ArrayList<MatchedInstance>>  irTreePattern = new Hashtable<Integer, ArrayList<MatchedInstance>>();
    private static  ArrayList<Statement> nonDetectedExpressions= new ArrayList<Statement>();
	
	public static ArrayList<MatchedInstance>  GetChildrensPattern(Traversable parentNode)
	{
		if(irTreePattern.containsKey(System.identityHashCode(parentNode)))
			return  irTreePattern.get(System.identityHashCode(parentNode));
		else 
			return null;
		
	}
	
	public static ArrayList<MatchedInstance> GetNodesPattern(Traversable currentNode )
	{
		if(irTreePattern.containsKey(System.identityHashCode(currentNode)))
			return irTreePattern.get(System.identityHashCode(currentNode));
		else 
			return null;
		
	}
	
	
	public static void  AddDetectedPattern(MatchedInstance newPattern,Traversable parentNode)
	{
		if(isPrunable(newPattern.getTreeNodeReference(),parentNode ))
			return;
		if(irTreePattern.containsKey(System.identityHashCode(parentNode)))
		{
			ArrayList<MatchedInstance> tmpPatternList = irTreePattern.get(System.identityHashCode(parentNode));
			tmpPatternList.add(newPattern);
			
			
		}
		else 
		{   
			ArrayList<MatchedInstance> tmpPatternList = new ArrayList<MatchedInstance>();
			tmpPatternList.add(newPattern);
			irTreePattern.put(System.identityHashCode(parentNode), tmpPatternList);
		}
		
		// for analysis purposes
		if(newPattern.getInstance().getPattern().getName().equals("No_Pattern"))
			nonDetectedExpressions.add((Statement)newPattern.getTreeNodeReference());
		
	}
	
	// 
	//inserts the newly detected pattern to the start of the list
	public static void  addDetectedPatternAtFirst(MatchedInstance newPattern,Traversable parentNode)
	{
		if(isPrunable(newPattern.getTreeNodeReference(),parentNode ))
			return;
		if(irTreePattern.containsKey(System.identityHashCode(parentNode)))
		{
			ArrayList<MatchedInstance> tmpPatternList = irTreePattern.get(System.identityHashCode(parentNode));
			tmpPatternList.add(0,newPattern);
			
		}
		else 
		{   
			ArrayList<MatchedInstance> tmpPatternList = new ArrayList<MatchedInstance>();
			tmpPatternList.add(newPattern);
			irTreePattern.put(System.identityHashCode(parentNode), tmpPatternList);
		}
		
		// for analysis purposes
		if(newPattern.getInstance().getPattern().getName().equals("No_Pattern"))
			nonDetectedExpressions.add((Statement)newPattern.getTreeNodeReference());
		
	}
	
	
	
	// since cetus provides some extra nodes in the hierarchy such as ExpressionStatement,CompundStatement ,sometimes it is necessary to
	//shift the node to the upper level
	public static void  ShiftPattern(Traversable sourceNode)
	{
		if(irTreePattern.containsKey(System.identityHashCode(sourceNode)))
		{
			ArrayList<MatchedInstance> tmpPatternList = irTreePattern.get(System.identityHashCode(sourceNode));
			irTreePattern.remove(System.identityHashCode(sourceNode));
			if(!isPrunable(sourceNode))
			for(int i= 0; i<tmpPatternList.size();i++)	
					if(!isFalselyDetected(tmpPatternList.get(i), sourceNode.getParent()))
					{
						AddDetectedPattern(tmpPatternList.get(i), sourceNode.getParent());
						RecognitionAnalysis.updateExpressionStatementInfo(sourceNode);
					}	
			
			
		}
		
	}
	
	// checks whether a node is the initialStatement or the step or the condition of a loop 
	public static boolean isPrunable(Traversable sourceNode)
	{
		
		if(sourceNode == null)
			return false;
		if(sourceNode.getParent() instanceof ForLoop)
		{
			ForLoop tmpLoop = (ForLoop) sourceNode.getParent();
			
			if( tmpLoop.getInitialStatement()!=null && (tmpLoop.getInitialStatement().equals(sourceNode))
			    ||
			    (tmpLoop.getCondition()!=null && tmpLoop.getCondition().equals(sourceNode))
			    ||
			    LoopFunctions.isStep(tmpLoop,sourceNode) )
				return true ;
			
			return false ;
		}
	/*	if(sourceNode.getParent() instanceof IfStatement)
		{
			IfStatement tmpStatement =(IfStatement)sourceNode.getParent();
			if(tmpStatement.getControlExpression()!=null && tmpStatement.getControlExpression().equals(sourceNode))
				return true ;
			return false ;
		}*/
		return false;
	}
	
	
	public static boolean isPrunable(Traversable sourceNode,Traversable parentNode)
	{
		if(parentNode instanceof ForLoop)
		{
			ForLoop tmpLoop = (ForLoop) parentNode;
			
			if( tmpLoop.getInitialStatement()!=null && (tmpLoop.getInitialStatement().equals(sourceNode))
			    ||
			    (tmpLoop.getCondition()!=null && tmpLoop.getCondition().equals(sourceNode))
			    ||
			    LoopFunctions.isStep(tmpLoop,sourceNode) )
				return true ;
			
			return false ;
		}
		return false;
	}
	
	// attempt to delete a pattern from a list of children pattern if it can be found 
	public static void deleteChildrenPattern(Traversable parentNode,Traversable childNode)
	{
		 ArrayList<MatchedInstance> childrenPattern= AnnotatedTree.GetChildrensPattern(parentNode);
		     int j= 0;
		      for(; j < childrenPattern.size();j++)
		    	  if(System.identityHashCode(childrenPattern.get(j).getTreeNodeReference())==System.identityHashCode(childNode))
		    		  break ;
		      if(j<childrenPattern.size())
		    	   childrenPattern.remove(j);
		
		    	  
	}

	
	
	public static boolean isFalselyDetected(MatchedInstance patternStructure, Traversable parentNode)
	{

	     Traversable node = patternStructure.getTreeNodeReference(); 
	     if(node instanceof Expression)
	    	 return isPrunable(node, parentNode);
	     
	     if( node instanceof ExpressionStatement)
	    	 
	    	 return  isPrunable(((ExpressionStatement)node).getExpression(),parentNode);
	     
	     return  false ;
		
		
		
	}
	

	public static void Clear()
	{
		
		irTreePattern.clear();
		nonDetectedExpressions.clear();
	
		
	}
	
	
	public  static ArrayList<Statement> getNonDetectedPattern()
	{
		return nonDetectedExpressions;
	}

}
