package ida.pelab.prt.patternrecognitionTool;

import ida.pelab.prt.pattern.MatchedInstance;
import ida.pelab.prt.pattern.PatternStructure;

import java.lang.reflect.Method;
import java.util.ArrayList;

import cetus.hir.Traversable;

/**
 * This class loads the user-provided constraint checking for the recognition process of a newly user-defined pattern.  
 */
public class AuxiliaryRecognitionModule {

	private static IPatternComparable  customUserConstraint;

	/**
	 * Passes the user-defined constraint class to the recognition module.
	 * @param customUserConstraint
	 */
	public static void setCustomUserConstraint(IPatternComparable customUserConstraint) {
		AuxiliaryRecognitionModule.customUserConstraint = customUserConstraint;
	}


	public static  boolean compare (PatternStructure pattern,Traversable subtreeRoot,ArrayList<MatchedInstance> patternList)
	{
		
		if(customUserConstraint==null)
			return false;
		try
		{
			
			Class<?>  auxiliaryClass= customUserConstraint.getClass();
			Class<?> []  parameterTypes= new Class[3];
			parameterTypes[0]=PatternStructure.class;
			parameterTypes[1]=Traversable.class;
			parameterTypes[2]=ArrayList.class;
			String methodName= pattern.getName()+"Compare";
			Method compareMethod =auxiliaryClass.getMethod(methodName, parameterTypes);
			Object [] arguments= new Object[3];
			arguments[0]=pattern;
			arguments[1]=subtreeRoot;
			arguments[2]=patternList;
			Object classObject =auxiliaryClass.newInstance();
			Object output=compareMethod.invoke(classObject, arguments);
		 return (Boolean)output;
			
		}
		
		catch(Exception e)
		{
			System.err.println("Can not find/load the custom constraint checking method for the \""+ pattern.getName() + "\" pattern");
			return false;
		}
	}

	
	
	
	
}
