package ida.pelab.prt.patternrecognitionTool;




import ida.pelab.prt.generalfunctions.IgnoredNodes;
import ida.pelab.prt.generalfunctions.TraverseFunction;




import ida.pelab.prt.pattern.MatchedInstance;
import ida.pelab.prt.pattern.PatternLevel;
import ida.pelab.prt.pattern.PatternStructure;
import ida.pelab.prt.pattern.PatternType;
import ida.pelab.prt.patternhandlers.MappedInstanceHandler;

import ida.pelab.prt.patternhierarchy.AnnotatedTree;
import ida.pelab.prt.patternhierarchy.PatternHierarchyGraph;

import java.util.ArrayList;



import cetus.hir.ArrayAccess;
import cetus.hir.CompoundStatement;
import cetus.hir.Statement;

import cetus.hir.Traversable;

public class MatchingModule {

	
	public void applyVerticalRecognition(Traversable treeNode) {
		
	    
		if(!(treeNode instanceof ArrayAccess) && TraverseFunction.hasChildren(treeNode) && ! IgnoredNodes.isUninteresting(treeNode))
		{
			ArrayList<MatchedInstance>  childrensPattern = AnnotatedTree.GetChildrensPattern(treeNode);
			if(childrensPattern!=null)
			{
				if(childrensPattern.size()>0)
					{	
					     
						if(IgnoredNodes.isMovable(treeNode))						
							AnnotatedTree.ShiftPattern(treeNode);
																		
						else 
						{
							//apply horizontal pattern recognition
						    //getting the candidate patterns in PGH that can be built from this pattern
							//BasePattern currentLevelPattern =childrensPattern.get(childrensPattern.size()-1).getDetectedPattern();

							PatternStructure currentLevelPattern =getMaximumChildLevel(childrensPattern);
							ArrayList<PatternStructure> nextLevelPattern = PatternHierarchyGraph.getTopLevelPattern(currentLevelPattern);
						//	nextLevelPattern =NewPatternHierarchyGraph.extendList(nextLevelPattern);
							if(nextLevelPattern!=null && nextLevelPattern.size()>0)
								{
									int i=0;
									for(; i< nextLevelPattern.size();i++)
											{
												PatternStructure patternInLevel =nextLevelPattern.get(i);
												if(!patternInLevel.getType().equals(PatternType.Horizontal))
													if(StructuralComparison.compare(patternInLevel, treeNode,childrensPattern))
														{
														handleDetectedPattern(patternInLevel, treeNode,childrensPattern);
														break;														
														}
											}
								// the  structure does not match with any of the current pattern	
							 if(i==nextLevelPattern.size() && treeNode instanceof Statement )
								 handleDetectedPattern(PatternStructure.getNoPatternInstance(), treeNode,null);
								}
							else 
							{     
								 // there are no top level patterns for this pattern
								if(treeNode instanceof Statement)	
									handleDetectedPattern(PatternStructure.getNoPatternInstance(), treeNode,null);
							}
						}
				    }
				
			
			
			}
			// there is an expression statement which  is not detected as a pattern
			else if(treeNode instanceof Statement)	
				handleDetectedPattern(PatternStructure.getNoPatternInstance(), treeNode,null);
		

		
			
			
	 }
	// leaf nodes ,they can be either identifier ,variable or  arrayAcess
		else 
		{
			ArrayList<PatternStructure> nextLevelPattern =PatternHierarchyGraph.getPatternsByLevel(PatternLevel.Trivial);
			if(nextLevelPattern!=null && nextLevelPattern.size()>0)
				for(int i=0; i< nextLevelPattern.size();i++)
					{
						PatternStructure patternInLevel =nextLevelPattern.get(i);						
						if(StructuralComparison.compare(patternInLevel,treeNode,null))
								{
								handleDetectedPattern(patternInLevel, treeNode,null);
								break;
				
								}
					}
			
		}
		// TODO Auto-generated method stub
		
	}
	
	
	
	
	
	
	public void handleDetectedPattern(PatternStructure detectedPattern,Traversable nodeLevel, ArrayList<MatchedInstance> subPatterns)
	{
			MappedInstanceHandler instanceHandler= new MappedInstanceHandler();
			instanceHandler.handle(detectedPattern, nodeLevel,subPatterns);
		
	}
	
	
	public PatternStructure  getMaximumChildLevel(ArrayList<MatchedInstance> childrenPattern)
	{
		
		if(childrenPattern.size()<1)
			return null;
		else 
		{
			PatternStructure result= childrenPattern.get(0).getInstance().getPattern();
			for(int i= 1;i<childrenPattern.size();i++)
			{
				PatternStructure nextNode=childrenPattern.get(i).getInstance().getPattern();				
				if(nextNode.getLevel().ordinal()>result.getLevel().ordinal())
				 result=nextNode;
			}
			
			return result;
		}
		
	}
	




	// for horizontal patterns built on top of another horizontal pattern at the same level, the horizontal pattern recognition
	//must be repeated
	// the function returns true for only those specific patterns
	// an example : DOCPRODUCTHRT which is built on top of  VSUMHRT and both are at the same level 
	
	public boolean applyHorizontalRecognition(Traversable treeNode) {
		
		boolean anyDetected =false;
		if(!(treeNode instanceof CompoundStatement))
			return false;
		
		ArrayList<MatchedInstance>  childrensPattern = AnnotatedTree.GetChildrensPattern(treeNode);
		if(childrensPattern==null || childrensPattern.size()<2)
			return false ;
		
		ArrayList<PatternStructure> horizontalPatternList = PatternHierarchyGraph.getHorizontalList();
		for(int i= 0; i< horizontalPatternList.size();i++)
		{
			PatternStructure hrtPattern = horizontalPatternList.get(i);
			
			if(StructuralComparison.compareHorizontalStructure(hrtPattern, treeNode, childrensPattern))
			 anyDetected=true;
			
		}
		
		return anyDetected;
	}

}
